@@ -59,7 +59,7 @@ class AggressiveTypeInferencer(NodeTransformer):
 
     def visit_ClassDef(self, node: ClassDef) -> ClassDef:
         class_record = RecordReader.extract(node)
-        self.set_variable_type(node.name, ClassType(class_record))
+        self.set_variable_type(node.name, RecordType(class_record))
         return node
 
     # TODO type inference for classDef
@@ -74,7 +74,7 @@ class AggressiveTypeInferencer(NodeTransformer):
         if tc.value is None:
             tc.typ = NoneType()
         else:
-            tc.typ = InstanceType(type(node.value).__name__)
+            tc.typ = RecordInstanceType(type(node.value).__name__)
         return tc
 
     def visit_Tuple(self, node: Tuple) -> TypedTuple:
@@ -132,11 +132,13 @@ class AggressiveTypeInferencer(NodeTransformer):
             )
         vartyp = None
         itertyp = typed_for.iter.typ
-        if isinstance(itertyp, ListType) or isinstance(itertyp, TupleType):
+        if isinstance(itertyp, TupleType):
             vartyp = itertyp.typs[0]
             assert all(
                 itertyp.typs[0] == t for t in typed_for.iter.typ.typs
-            ), "Iterating through a list requires the same type for each element"
+            ), "Iterating through a tuple requires the same type for each element"
+        elif isinstance(itertyp, ListType):
+            vartyp = itertyp.typ
         else:
             raise NotImplementedError(
                 "Type inference for non-list objects is not supported"
@@ -233,10 +235,36 @@ class AggressiveTypeInferencer(NodeTransformer):
 
     def visit_Subscript(self, node: Subscript) -> TypedSubscript:
         ts = copy(node)
-        ts.value = self.visit(node.value)
         assert isinstance(
             ts.slice, Index
         ), "Only single index slices are currently supported"
+        # special case: Subscript of Union / Optional
+        if isinstance(ts.value, Name) and ts.value.id in ["Union", "Optional"]:
+            if ts.value.id == "Union":
+                assert isinstance(
+                    ts.slice.value, Tuple
+                ), "Union must combine multiple classes"
+                typs = self.visit(node.slice.value)
+                eltyps = [e.typ for e in typs.elts]
+                assert all(
+                    isinstance(e, ClassType) for e in eltyps
+                ), "Union must combine classes"
+                ts.value = UnionType(eltyps)
+                ts.typ = UnionType(eltyps)
+                return ts
+            if ts.value.id == "Optional":
+                assert isinstance(
+                    ts.slice.value, Name
+                ), "Optional must have a single type as parameter"
+                e = self.visit(node.slice.value)
+                assert isinstance(
+                    e.typ, ClassType
+                ), "Optional must have a type as parameter"
+                ts.value = OptionalType(e)
+                ts.typ = OptionalType(e)
+                return ts
+
+        ts.value = self.visit(node.value)
         if isinstance(ts.value.typ, TupleType) or isinstance(ts.value.typ, ListType):
             if all(ts.value.typ.typs[0] == t for t in ts.value.typ.typs):
                 ts.typ = ts.value.typ.typs[0]
@@ -278,13 +306,16 @@ class AggressiveTypeInferencer(NodeTransformer):
         tp = copy(node)
         tp.value = self.visit(node.value)
         owner = tp.value.typ
-        assert isinstance(owner, InstanceType), "Accessing attribute of non-instance"
+        assert isinstance(
+            owner, RecordInstanceType
+        ), "Accessing attribute of non-instance"
         # look up class type in local scope
         owner_typ = self.variable_type(owner.typ)
         assert isinstance(
-            owner_typ, ClassType
+            owner_typ, RecordType
         ), "Accessing attribute of instance of a non-class"
         tp.typ = None
+        # TODO rewrite access to constructor
         for i, (attr_name, attr_type) in enumerate(owner_typ.record.fields):
             if attr_name == tp.attr:
                 tp.typ = attr_type
